"""
Generators which yield an id to include in a JSON-RPC request.

By default the request `id` is a decimal number which increments with each request. See
the `config` module.
"""
import itertools
from random import choice
from string import ascii_lowercase, digits
from typing import Iterator
from uuid import uuid4


def decimal(start: int = 1) -> Iterator[int]:
    """
    Increments from `start`.

    e.g. 1, 2, 3, .. 9, 10, 11, etc.

    Args:
        start: The first value to start with.
    """
    return itertools.count(start)


def hexadecimal(start: int = 1) -> Iterator[str]:
    """
    Incremental hexadecimal numbers.

    e.g. 1, 2, 3, .. 9, a, b, etc.

    Args:
        start: The first value to start with.
    """
    while True:
        yield "%x" % start
        start += 1


def random(length: int = 8, chars: str = digits + ascii_lowercase) -> Iterator[str]:
    """
    A random string.

    Not unique, but has around 1 in a million chance of collision (with the default 8
    character length). e.g. 'fubui5e6'

    Args:
        length: Length of the random string.
        chars: The characters to randomly choose from.
    """
    while True:
        yield "".join([choice(chars) for _ in range(length)])


def uuid() -> Iterator[str]:
    """
    Unique uuid ids.

    For example, '9bfe2c93-717e-4a45-b91b-55422c5af4ff'
    """
    while True:
        yield str(uuid4())



"""
Low-level socket client.
"""
import socket
from typing import Any

from ..client import Client
from ..response import Response


class SocketClient(Client):
    """
    Args:
        socket: Connected socket.
        encoding: The charset to encode and decode the data with.
        delimiter: String marking the end of a request or response.
        *args: Passed through to Client class.
        **kwargs: Passed through to Client class.
    """

    def __init__(
        self,
        socket: socket.socket,
        *args: Any,
        encoding: str = "utf-8",
        delimiter: str = "\n",
        **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.socket = socket
        self.delimiter = delimiter
        self.encoding = encoding
        self.delimiter_length = len(delimiter)

    def send_message(self, request: str, **kwargs: Any) -> Response:
        """
        Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :return: The response (a string for requests, None for notifications).
        """
        payload = str(request) + self.delimiter
        self.socket.send(payload.encode(self.encoding))

        response = bytes()
        decoded = None

        # Receive the response until we find the delimiter.
        # TODO Do not wait for a response if the message sent is a notification.
        while True:
            response += self.socket.recv(1024)

            decoded = response.decode(self.encoding)
            if len(decoded) < self.delimiter_length:
                continue

            # TODO Check that're not in the middle of the response.
            elif decoded[-self.delimiter_length :] == self.delimiter:
                break

        assert decoded is not None
        return Response(decoded[: -self.delimiter_length])



"""
Config object.

Allows modules to import a pre-loaded configuration object.
"""
import importlib
import os
from configparser import ConfigParser
from typing import Iterator


def parse_callable(path: str) -> Iterator:
    """
    ConfigParser converter.

    Calls the specified object, e.g. Option "id_generators.decimal" returns
    `id_generators.decimal()`.
    """
    module = path[: path.rindex(".")]
    callable_name = path[path.rindex(".") + 1 :]
    callable_ = getattr(importlib.import_module(module), callable_name)
    return callable_()


defaults = {
    "trim_log_values": "False",
    "validate_against_schema": "True",
    "id_generator": "jsonrpcclient.id_generators.decimal",
}
config = ConfigParser(
    defaults=defaults,
    default_section="general",
    converters={"callable": parse_callable},
)
config.read([".jsonrpcclientrc", os.path.expanduser("~/.jsonrpcclientrc")])



"""
Websockets client.

http://websockets.readthedocs.io/
"""
from typing import Any

from websockets import WebSocketCommonProtocol  # type: ignore

from ..async_client import AsyncClient
from ..response import Response


class WebSocketsClient(AsyncClient):
    def __init__(
        self, socket: WebSocketCommonProtocol, *args: Any, **kwargs: Any
    ) -> None:
        """
        Args:
            socket: Connected websocket (websockets.connect("ws://localhost:5000"))
        """
        super().__init__(*args, **kwargs)
        self.socket = socket

    async def send_message(self, request: str, **kwargs: Any):  # type: ignore
        await self.socket.send(request)
        response_text = await self.socket.recv()
        return Response(response_text)



"""Logging"""
import json
import logging
from typing import Any, Dict, List, Optional, Union, cast


def _trim_string(message: str) -> str:
    longest_string = 30

    if len(message) > longest_string:
        prefix_len = int(longest_string / 3)
        suffix_len = prefix_len
        return message[:prefix_len] + "..." + message[-suffix_len:]

    return message


def _trim_dict(message_obj: Dict[str, Any]) -> Dict[str, Any]:
    result = {}
    longest_list = 30
    for k, val in message_obj.items():
        if isinstance(val, str):
            result[k] = _trim_string(val)
        elif isinstance(val, list) and len(val) > longest_list:
            prefix_len = int(longest_list / 3)
            suffix_len = prefix_len
            result[k] = cast(str, val[:prefix_len] + ["..."] + val[-suffix_len:])
        elif isinstance(val, dict):
            result[k] = cast(str, _trim_values(val))
        else:
            result[k] = val
    return result


def _trim_values(message_obj: Union[Dict, List]) -> Union[Dict, List]:
    # Batch?
    if isinstance(message_obj, list):
        return [_trim_dict(i) for i in message_obj]
    else:
        return _trim_dict(message_obj)


def _trim_message(message: str) -> str:
    # Attempt to deserialize
    try:
        message_obj = json.loads(message)
    except ValueError:
        # Could not be deserialized, trim the string anyway.
        return _trim_string(str(message))
    else:
        return json.dumps(_trim_values(message_obj))


def log_(
    message: str,
    logger: logging.Logger,
    level: str = "info",
    extra: Optional[Dict] = None,
    trim: bool = False,
) -> None:
    """
    Log a request or response

    Args:
        message: JSON-RPC request or response string.
        level: Log level.
        extra: More details to include in the log entry.
        trim: Abbreviate log messages.
    """
    if extra is None:
        extra = {}
    # Clean up the message for logging
    if message:
        message = message.replace("\n", "").replace("  ", " ").replace("{ ", "{")
    if trim:
        message = _trim_message(message)
    # Log.
    getattr(logger, level)(message, extra=extra)



# Convenience functions
from jsonrpcclient.response import Response


def request(endpoint: str, *args: list, **kwargs: dict) -> Response:
    from .clients.http_client import HTTPClient
    return HTTPClient(endpoint).request(*args, **kwargs)


def notify(endpoint: str, *args: list, **kwargs: dict) -> Response:
    from .clients.http_client import HTTPClient
    return HTTPClient(endpoint).request(*args, **kwargs)


def send(endpoint: str, *args: list, **kwargs: dict) -> Response:
    from .clients.http_client import HTTPClient
    return HTTPClient(endpoint).send(*args, **kwargs)



"""
Response and JSONRPCResponse classes.

This module needs a major overhaul.
"""
from collections import OrderedDict
from json import dumps as serialize
from typing import Any, Dict, List, Union

NOID = object()


def sort_response(response: Dict[str, Any]) -> OrderedDict:
    """
    Sort the keys in a JSON-RPC response object.

    This has no effect other than making it nicer to read. Useful in Python 3.5 only,
    dictionaries are already sorted in newer Python versions.

    Example::

        >>> json.dumps(sort_response({'id': 2, 'result': 5, 'jsonrpc': '2.0'}))
        {"jsonrpc": "2.0", "result": 5, "id": 1}

    Args:
        response: Deserialized JSON-RPC response.

    Returns:
        The same response, sorted in an OrderedDict.
    """
    root_order = ["jsonrpc", "result", "error", "id"]
    error_order = ["code", "message", "data"]
    req = OrderedDict(sorted(response.items(), key=lambda k: root_order.index(k[0])))
    if "error" in response:
        req["error"] = OrderedDict(
            sorted(response["error"].items(), key=lambda k: error_order.index(k[0]))
        )
    return req


class JSONRPCResponse:
    """
    A parsed JSON-RPC response object.

    Base class for the responses. There should be no need to validate the input data to
    these responses, since the data hass passed the jsonschema validation.
    """

    ok = False

    def __init__(self, jsonrpc: str, id: Any) -> None:
        self.jsonrpc = jsonrpc
        self.id = id


class SuccessResponse(JSONRPCResponse):
    """
    Represents a JSON-RPC success response object.
    """

    ok = True

    def __init__(self, result: Any, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.result = result

    def __repr__(self) -> str:
        return "<SuccessResponse(id={}, result={})>".format(self.id, self.result)

    def __str__(self) -> str:
        return serialize(
            sort_response(dict(jsonrpc=self.jsonrpc, result=self.result, id=self.id))
        )


class NotificationResponse(SuccessResponse):
    """
    Represents a JSON-RPC notification response object.
    """

    ok = True

    def __init__(self) -> None:
        super().__init__(jsonrpc="2.0", result=None, id=NOID)

    def __repr__(self) -> str:
        return "<NotificationResponse()>"

    def __str__(self) -> str:
        return ""


class ErrorResponse(JSONRPCResponse):
    """
    Represents a JSON-RPC error response object.
    """

    ok = False

    def __init__(self, error: Dict[str, Any], **kwargs: Any) -> None:
        super().__init__(id=kwargs.pop("id", NOID), **kwargs)
        self.message = error.get("message")
        self.code = error.get("code")
        self.data = error.get("data")

    def __repr__(self) -> str:
        if self.id is NOID:
            return '<ErrorResponse(message="{}")>'.format(self.message)
        return '<ErrorResponse(id={}, message="{}")>'.format(self.id, self.message)

    def __str__(self) -> str:
        error = dict(code=self.code, message=self.message)
        if self.data:
            error["data"] = self.data
        deserialized = dict(jsonrpc=self.jsonrpc, error=error)
        if self.id is not NOID:
            deserialized["id"] = self.id
        return serialize(sort_response(deserialized))


def total_results(
    data: Union[List[JSONRPCResponse], JSONRPCResponse, None], *, ok: bool = True
) -> int:
    """
    Returns the total parsed responses, given the return value from parse().
    """
    if isinstance(data, list):
        return sum([1 for d in data if d.ok == ok])
    elif isinstance(data, JSONRPCResponse):
        return int(data.ok == ok)
    return 0  # The data hasn't been parsed yet. The data attribute hasn't been set.


class Response:
    """
    Wraps a client response.

    >>> Response(response.text, raw=response)
    """

    def __init__(self, text: str, raw: Any = None) -> None:
        """
        Args:
            text: The response string, as it was returned from the server.
            raw: The framework's own response object. Gives the user access to the
                framework (e.g. Requests library's `Response` object). (optional)
        """
        self.text = text
        self.raw = raw
        # Data is the parsed version of the response.
        self.data = None  # type: Union[JSONRPCResponse, List[JSONRPCResponse], None]

    def __repr__(self) -> str:
        total_ok = total_results(self.data, ok=True)
        total_errors = total_results(self.data, ok=False)
        if total_errors:
            return "<Response[{} ok, {} errors]>".format(total_ok, total_errors)
        return "<Response[{}]>".format(total_ok)



"""Parse response text, returning JSONRPCResponse objects."""
from json import loads as deserialize
from typing import Any, Dict, List, Union

import jsonschema  # type: ignore
from pkg_resources import resource_string

from .response import (ErrorResponse, JSONRPCResponse, NotificationResponse,
                       SuccessResponse)

schema = deserialize(resource_string(__name__, "response-schema.json").decode())


def get_response(response: Dict[str, Any]) -> JSONRPCResponse:
    """
    Converts a deserialized response into a JSONRPCResponse object.

    The dictionary be either an error or success response, never a notification.

    Args:
        response: Deserialized response dictionary. We can assume the response is valid
            JSON-RPC here, since it passed the jsonschema validation.
    """
    if "error" in response:
        return ErrorResponse(**response)
    return SuccessResponse(**response)


def parse(
    response_text: str, *, batch: bool, validate_against_schema: bool = True
) -> Union[JSONRPCResponse, List[JSONRPCResponse]]:
    """
    Parses response text, returning JSONRPCResponse objects.

    Args:
        response_text: JSON-RPC response string.
        is_batch_request: If the response_text is an empty string, this determines how
            to parse.
        validate_against_schema: Validate against the json-rpc schema.

    Returns:
        Either a JSONRPCResponse, or a list of them.

    Raises:
        json.JSONDecodeError: The response was not valid JSON.
        jsonschema.ValidationError: The response was not a valid JSON-RPC response
            object.
    """
    # If the response is empty, we can't deserialize it; an empty string is valid
    # JSON-RPC, but not valid JSON.
    if not response_text:
        if batch:
            # An empty string is a valid response to a batch request, when there were
            # only notifications in the batch.
            return []
        else:
            # An empty string is valid response to a Notification request.
            return NotificationResponse()

    # If a string, ensure it's json-deserializable
    deserialized = deserialize(response_text)

    # Validate the response against the Response schema (raises
    # jsonschema.ValidationError if invalid)
    if validate_against_schema:
        jsonschema.validate(deserialized, schema)

    # Batch response
    if isinstance(deserialized, list):
        return [get_response(r) for r in deserialized if "id" in r]
    # Single response
    return get_response(deserialized)



"""
Client class.

Base class for the clients.
"""
import logging
from abc import ABCMeta, abstractmethod
from json import dumps as serialize
from typing import Any, Callable, Dict, Iterator, List, Optional, Union

from apply_defaults import apply_config, apply_self  # type: ignore

from .config import config
from .exceptions import ReceivedErrorResponseError
from .log import log_
from .parse import parse
from .requests import Notification, Request
from .response import ErrorResponse, Response

request_log = logging.getLogger(__name__ + ".request")
response_log = logging.getLogger(__name__ + ".response")


def is_batch_request(request_text: str) -> bool:
    try:
        return request_text.strip()[0] == "["
    except IndexError:
        return False


class Client(metaclass=ABCMeta):
    """
    Protocol-agnostic base class for clients.

    Subclasses must override `send_message` to transport the message.
    """

    DEFAULT_REQUEST_LOG_FORMAT = "--> %(message)s"
    DEFAULT_RESPONSE_LOG_FORMAT = "<-- %(message)s"

    @apply_config(config, converters={"id_generator": "getcallable"})
    def __init__(
        self,
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        id_generator: Optional[Iterator] = None,
        basic_logging: bool = False,
    ) -> None:
        """
        Args:
            trim_log_values: Abbreviate the log entries of requests and responses.
            validate_against_schema: Validate response against the JSON-RPC schema.
            id_generator: Iterable of values to use as the "id" part of the request.
            basic_logging: Will create log handlers to output request & response
                messages.
        """
        self.trim_log_values = trim_log_values
        self.validate_against_schema = validate_against_schema
        self.id_generator = id_generator
        if basic_logging:
            self.basic_logging()

    def basic_logging(self) -> None:
        """
        Call this on the client object to create log handlers to output request and
        response messages.
        """
        # Request handler
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt=self.DEFAULT_REQUEST_LOG_FORMAT))
        request_log.addHandler(handler)
        request_log.setLevel(logging.INFO)
        # Response handler
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt=self.DEFAULT_RESPONSE_LOG_FORMAT))
        response_log.addHandler(handler)
        response_log.setLevel(logging.INFO)

    @apply_self
    def log_request(
        self, request: str, trim_log_values: bool = False, **kwargs: Any
    ) -> None:
        """
        Log a request.

        Args:
            request: The JSON-RPC request string.
            trim_log_values: Log an abbreviated version of the request.
        """
        return log_(request, request_log, "info", trim=trim_log_values, **kwargs)

    @apply_self
    def log_response(
        self, response: Response, trim_log_values: bool = False, **kwargs: Any
    ) -> None:
        """
        Log a response.

        Note this is different to log_request, in that it takes a Response object, not a
        string.

        Args:
            response: The Response object to log. Note this is different to log_request
                which takes a string.
            trim_log_values: Log an abbreviated version of the response.
        """
        return log_(response.text, response_log, "info", trim=trim_log_values, **kwargs)

    @abstractmethod
    def send_message(self, request: str, **kwargs: Any) -> Response:
        """
        Transport the request to the server.

        Override this method in the protocol-specific subclasses.

        Args:
            request: A JSON-RPC request.
        """

    def validate_response(self, response: Response) -> None:
        """
        Can be overridden for custom validation of the response.

        Raise an exception to fail validation.
        """
        pass

    @apply_self
    def send(
        self,
        request: Union[str, Dict, List],
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        **kwargs: Any
    ) -> Response:
        """
        Send a request, passing the whole JSON-RPC request object.

        After sending, logs, validates and parses.

        >>> client.send('{"jsonrpc": "2.0", "method": "ping", "id": 1}')
        <Response[1]>

        Args:
            request: The JSON-RPC request. Can be either a JSON-encoded string or a
                Request/Notification object.
            trim_log_values: Abbreviate the log entries of requests and responses.
            validate_against_schema: Validate response against the JSON-RPC schema.
            kwargs: Clients can use this to configure an single request. For example,
                HTTPClient passes this through to `requests.Session.send()`.
            in the case of a Notification.
        """
        # Convert the request to a string if it's not already.
        request_text = request if isinstance(request, str) else serialize(request)
        batch = is_batch_request(request_text)
        self.log_request(request_text, trim_log_values=trim_log_values)
        response = self.send_message(request_text, **kwargs)
        self.log_response(response, trim_log_values=trim_log_values)
        self.validate_response(response)
        response.data = parse(
            response.text, batch=batch, validate_against_schema=validate_against_schema
        )
        # If received a single error response, raise
        if isinstance(response.data, ErrorResponse):
            raise ReceivedErrorResponseError(response.data)
        return response

    @apply_self
    def notify(
        self,
        method_name: str,
        *args: Any,
        trim_log_values: Optional[bool] = None,
        validate_against_schema: Optional[bool] = None,
        **kwargs: Any
    ) -> Response:
        """
        Send a JSON-RPC request, without expecting a response.

        Args:
            method_name: The remote procedure's method name.
            args: Positional arguments passed to the remote procedure.
            kwargs: Keyword arguments passed to the remote procedure.
            trim_log_values: Abbreviate the log entries of requests and responses.
            validate_against_schema: Validate response against the JSON-RPC schema.
        """
        return self.send(
            Notification(method_name, *args, **kwargs),
            trim_log_values=trim_log_values,
            validate_against_schema=validate_against_schema,
        )

    @apply_self
    def request(
        self,
        method_name: str,
        *args: Any,
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        id_generator: Optional[Iterator] = None,
        **kwargs: Any
    ) -> Response:
        """
        Send a request by passing the method and arguments.

        >>> client.request("cat", name="Yoko")
        <Response[1]

        Args:
            method_name: The remote procedure's method name.
            args: Positional arguments passed to the remote procedure.
            kwargs: Keyword arguments passed to the remote procedure.
            trim_log_values: Abbreviate the log entries of requests and responses.
            validate_against_schema: Validate response against the JSON-RPC schema.
            id_generator: Iterable of values to use as the "id" part of the request.
        """
        return self.send(
            Request(method_name, id_generator=id_generator, *args, **kwargs),
            trim_log_values=trim_log_values,
            validate_against_schema=validate_against_schema,
        )

    def __getattr__(self, name: str) -> Callable:
        """
        This gives us an alternate way to make a request.

        >>> client.cube(3)
        --> {"jsonrpc": "2.0", "method": "cube", "params": [3], "id": 1}

        That's the same as saying `client.request("cube", 3)`.
        """

        def attr_handler(*args: Any, **kwargs: Any) -> Response:
            return self.request(name, *args, **kwargs)

        return attr_handler



"""
aiohttp client.

Requires aiohttp >= 3.0.
http://aiohttp.readthedocs.io/
"""
from ssl import SSLContext
from typing import Any, Optional

import async_timeout  # type: ignore
from aiohttp import ClientSession  # type: ignore

from ..async_client import AsyncClient
from ..exceptions import ReceivedNon2xxResponseError
from ..response import Response


class AiohttpClient(AsyncClient):

    DEFAULT_RESPONSE_LOG_FORMAT = "<-- %(message)s (%(http_code)s %(http_reason)s)"

    def __init__(
        self,
        session: ClientSession,
        endpoint: str,
        *args: Any,
        ssl: Optional[SSLContext] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.endpoint = endpoint
        self.session = session
        self.ssl = ssl

    def log_response(
        self, response: Response, trim_log_values: bool = False, **kwargs: Any
    ) -> None:
        extra = (
            {"http_code": response.raw.status, "http_reason": response.raw.reason}
            if response.raw is not None
            else {}
        )
        super().log_response(
            response, extra=extra, trim_log_values=trim_log_values, **kwargs
        )

    def validate_response(self, response: Response) -> None:
        if response.raw is not None and not 200 <= response.raw.status <= 299:
            raise ReceivedNon2xxResponseError(response.raw.status)

    async def send_message(self, request: str) -> Response:  # type: ignore
        with async_timeout.timeout(10):
            async with self.session.post(
                self.endpoint, data=request, ssl=self.ssl
            ) as response:
                response_text = await response.text()
                return Response(response_text, raw=response)



"""
Classes to help create JSON-RPC Request objects.

Named plural to distinguish it from the request convenience function.

To create a request:

    >>> Request("cat", name="Yoko")
    {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Yoko'}, 'id': 1}
"""
import json
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterator, List, Optional

from . import id_generators


def sort_request(request: Dict[str, Any]) -> OrderedDict:
    """
    Sort a JSON-RPC request dict.

    This has no effect other than making the request nicer to read.

        >>> json.dumps(sort_request(
        ...     {'id': 2, 'params': [2, 3], 'method': 'add', 'jsonrpc': '2.0'}))
        '{"jsonrpc": "2.0", "method": "add", "params": [2, 3], "id": 2}'

    Args:
        request: JSON-RPC request in dict format.
    """
    sort_order = ["jsonrpc", "method", "params", "id"]
    return OrderedDict(sorted(request.items(), key=lambda k: sort_order.index(k[0])))


class _RequestClassType(type):
    """
    Request Metaclass.

    Catches undefined attributes on the class.
    """

    def __getattr__(cls: Callable, name: str) -> Callable:
        """
        This gives us an alternate way to make a request:

        >>> Request.cat()
        {'jsonrpc': '2.0', 'method': 'cat', 'id': 1}

        That's the same as saying `Request("cat")`.
        """

        def attr_handler(*args: Any, **kwargs: Any) -> "Request":
            return cls(name, *args, **kwargs)

        return attr_handler


class Notification(dict, metaclass=_RequestClassType):  # type: ignore
    """
    A request which does not expect a response.

    >>> Notification("cat")
    {'jsonrpc': '2.0', 'method': 'cat'}

    The first argument is the *method*; everything else is *arguments* to the
    method:

    >>> Notification("cat", 'Yoko', 5)
    {'jsonrpc': '2.0', 'method': 'cat', params: ['Yoko', 5]}

    Keyword arguments are also acceptable:

    >>> Notification("cat", name="Yoko", age=5)
    {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Yoko', 'age': 5}}

    If you prefer, call the method as though it was a class attribute:

    >>> Notification.cat(name="Yoko", age=5)
    {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Yoko', 'age': 5}}

    Args:
        method: The method name.
        args: Positional arguments.
        kwargs: Keyword arguments.

    Returns:
        The JSON-RPC request in dictionary form.
    """

    def __init__(self, method: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(jsonrpc="2.0", method=method)
        # Build the 'params' part. Merge the positional and keyword arguments into one
        # list.
        params_list = []  # type: List
        if args:
            params_list.extend(args)
        if kwargs:
            params_list.append(kwargs)
        # Add the params to self.
        if params_list:
            # The 'params' can be either "by-position" (a list) or "by-name" (a dict).
            # If there's only one list or dict in the params list, take it out of the
            # enclosing list, ie. [] instead of [[]], {} instead of [{}].
            if len(params_list) == 1 and (isinstance(params_list[0], (dict, list))):
                self.update(params=params_list[0])
            else:
                self.update(params=params_list)

    def __str__(self) -> str:
        """Wrapper around request, returning a string instead of a dict"""
        return json.dumps(sort_request(self))


class Request(Notification):
    """
    Create a JSON-RPC request object
    http://www.jsonrpc.org/specification#request_object.

    >>> Request("cat", name="Yoko")
    {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Yoko'}, 'id': 1}

    Args:
        method: The `method` name.
        args: Positional arguments added to `params`.
        kwargs: Keyword arguments added to `params`. Use request_id=x to force the
            `id` to use.

    Returns:
        The JSON-RPC request in dictionary form.
    """

    id_generator = id_generators.decimal()

    def __init__(
        self,
        method: str,
        *args: Any,
        id_generator: Optional[Iterator[Any]] = None,
        **kwargs: Any
    ) -> None:
        # If 'request_id' is passed, use the specified id
        if "request_id" in kwargs:
            id_ = kwargs.pop("request_id", None)
        else:  # Get the next id from the generator
            id_generator = (
                id_generator if id_generator is not None else self.id_generator
            )
            id_ = next(id_generator)
        # We call super last, after popping the request_id
        super().__init__(method, *args, **kwargs)
        self.update(id=id_)



"""
ZeroMQ client.

ZMQClient('tcp://hostname:5555').request('go')
"""
from typing import Any

import zmq  # type: ignore

from ..client import Client
from ..response import Response


class ZeroMQClient(Client):
    def __init__(
        self, endpoint: str, *args: Any, socket_type: int = zmq.REQ, **kwargs: Any
    ) -> None:
        """
        Args:
            endpoint: The server address.
            socket_type: The zeromq socket type.
        """
        super().__init__(*args, **kwargs)
        self.context = zmq.Context()
        self.socket = self.context.socket(socket_type)
        self.socket.connect(endpoint)

    def send_message(self, request: str, **kwargs: Any) -> Response:
        """
        Transport the message to the server and return the response.

        Args:
            request: The JSON-RPC request string.

        Returns:
            A Response.
        """
        self.socket.send_string(request)
        return Response(self.socket.recv().decode())
